from __future__ import annotations
from typing import Any, TYPE_CHECKING
import numpy as np
from formable.levenberg_marquardt import (
    FittingParameter,
    LMFitterOptimisation,
    LMFitter,
)
from formable.tensile_test import TensileTest

if TYPE_CHECKING:
    from numpy.typing import NDArray
    from matflow.param_classes.single_crystal_parameters import SingleCrystalParameters


def fit_single_crystal_parameters(
    VE_response: dict,
    single_crystal_parameters: dict,
    tensile_test: dict,
    initial_damping: list[float] | None = None,
) -> dict[str, Any]:
    """Perform Levenberg-Marquardt optimisation."""

    # Generate FittingParameter objects:
    fitting_params = []
    null_perturbation_idx = None
    params: SingleCrystalParameters
    for idx, params in enumerate(single_crystal_parameters["iteration_0"]["value"]):
        if params.perturbations:
            perturb = params.perturbations[0]
            path = perturb["path"]
            name = "__".join([str(i) for i in path])
            value = _get_by_path(params.base, path)

            fitting_param_i = FittingParameter(
                name=name,
                values=[value],
                address=path,
                perturbation=perturb["multiplicative"],
            )
            fitting_params.append(fitting_param_i)
        else:
            # The null-perturbation does not correspond to a FittingParameter:
            null_perturbation_idx = idx

    # Collect volume element responses:
    tensile_tests_by_iteration = []
    for iteration_idx, all_vol_elem_resp in VE_response.items():
        # take x-normal component of tensors; comparable to experimental data
        # TODO: allow for non-x-direction and more generally rotated load cases
        tensile_tests = []
        for vol_elem_resp in all_vol_elem_resp["value"]:
            true_stress_tensor = vol_elem_resp["volume_data"]["vol_avg_stress"]["data"]
            true_strain_tensor = vol_elem_resp["volume_data"]["vol_avg_strain"]["data"]
            tensile_tests.append(
                TensileTest(
                    true_stress=true_stress_tensor[..., 0, 0],
                    true_strain=true_strain_tensor[..., 0, 0],
                )
            )

        # Need to reorder if null-perturbation is not first:
        num_sims_per_iteration = len(single_crystal_parameters["iteration_0"]["value"])
        if null_perturbation_idx != 0:
            non_null_pert_idx = list(
                set(range(num_sims_per_iteration)) - {null_perturbation_idx}
            )
            tensile_tests = [tensile_tests[null_perturbation_idx]] + [
                tensile_tests[i] for i in non_null_pert_idx
            ]

        tensile_tests_by_iteration.append(tensile_tests)

    # Generate fitter object:
    base_params = single_crystal_parameters["iteration_0"]["value"][0].base
    lm_fitter = LMFitter(
        exp_tensile_test=TensileTest(**tensile_test),
        single_crystal_parameters=base_params,
        fitting_params=fitting_params,
        initial_damping=initial_damping,
    )

    # Add simulated tests:
    for tensile_tests in tensile_tests_by_iteration:
        lm_fitter.add_simulated_tensile_tests(tensile_tests)

    optimised_single_crystal_parameters = lm_fitter.get_new_single_crystal_params(-1)

    outputs = {
        "single_crystal_parameters": {"phases": optimised_single_crystal_parameters},
        "levenberg_marquardt_fitter": lm_fitter.to_dict(),
    }
    return outputs


def _get_by_path(root: list | dict, path: list) -> Any:
    """Get a nested dict or list item according to its "key path"

    Parameters
    ----------
    root : dict or list
        Can be arbitrarily nested.
    path : list of str
        The address of the item to get within the `root` structure.

    Returns
    -------
    sub_data : any
    """

    sub_data: Any = root
    for key in path:
        sub_data = sub_data[key]

    return sub_data



import matflow as mf
from damask_parse.utils import validate_orientations


def import_VE_orientations(
    workflow_path, iteration_ID, task_name, element_idx, iteration_idx, input_or_output
):
    wk = mf.Workflow(workflow_path)
    if iteration_ID is not None:
        parent = wk.get_element_iterations_from_IDs([iteration_ID])[0]
    else:
        task = getattr(wk.tasks, task_name)
        parent = task.elements[element_idx]
        if iteration_idx is not None:
            parent = parent.iterations[iteration_idx]

    if input_or_output == "input":
        oris = parent.inputs.volume_element.value["orientations"]
    elif input_or_output == "output":
        oris = parent.outputs.volume_element.value["orientations"]

    # convert Zarr arrays to numpy arrays (TODO: should be able to encode Zarr arrays directly!)

    # see `LatticeDirection` enum:
    align_lookup = {
        "a": "A",
        "b": "B",
        "c": "C",
        "a*": "A_STAR",
        "b*": "B_STAR",
        "c*": "C_STAR",
    }
    unit_cell_alignment = {
        "x": align_lookup[oris["unit_cell_alignment"]["x"]],
        "y": align_lookup[oris["unit_cell_alignment"]["y"]],
        "z": align_lookup[oris["unit_cell_alignment"]["z"]],
    }
    type_lookup = {
        "quat": "QUATERNION",
        "euler": "EULER",
    }
    type_ = type_lookup[oris["type"]]

    oris_repr = {"type": type_}
    if type_ == "QUATERNION":
        oris_repr["quat_order"] = (
            oris["quat_component_ordering"].upper().replace("-", "_")
        )
    elif type_ == "EULER":
        oris_repr["euler_is_degrees"] = oris["euler_degrees"]

    oris = {
        "data": oris["quaternions"][:],
        "unit_cell_alignment": unit_cell_alignment,
        "representation": oris_repr,
    }
    return {"orientations": oris}



import h5py
import numpy as np

from damask_parse.utils import validate_volume_element


def parse_dream_3D_volume_element(dream_3D_hdf5_file: str):
    with h5py.File(dream_3D_hdf5_file, mode="r") as fh:
        synth_vol = fh["DataContainers"]["SyntheticVolumeDataContainer"]
        grid_size = synth_vol["_SIMPL_GEOMETRY"]["DIMENSIONS"][()]
        resolution = synth_vol["_SIMPL_GEOMETRY"]["SPACING"][()]
        size = [i * j for i, j in zip(resolution, grid_size)]

        # make zero-indexed:
        # (not sure why FeatureIds is 4D?)
        element_material_idx = synth_vol["CellData"]["FeatureIds"][()][..., 0] - 1
        element_material_idx = element_material_idx.transpose((2, 1, 0))

        num_grains = element_material_idx.max() + 1
        phase_names = synth_vol["CellEnsembleData"]["PhaseName"][()][1:]
        constituent_phase_idx = synth_vol["Grain Data"]["Phases"][()][1:] - 1
        constituent_phase_label = [
            phase_names[i][0].decode() for i in constituent_phase_idx
        ]
        eulers = synth_vol["Grain Data"]["EulerAngles"][()][1:]

    vol_elem = {
        "grid_size": grid_size,
        "size": size,
        "element_material_idx": element_material_idx,
        "constituent_material_idx": np.arange(num_grains),
        "constituent_phase_label": constituent_phase_label,
        "material_homog": ["SX"] * num_grains,
        "orientations": _process_dream3D_euler_angles(eulers),
    }
    return validate_volume_element(vol_elem)


def _process_dream3D_euler_angles(euler_angles, degrees: bool = False) -> dict:
    orientations = {
        "type": "euler",
        "euler_degrees": degrees,
        "euler_angles": euler_angles,
        "unit_cell_alignment": {"x": "a"},
    }
    return orientations



import numpy as np
from scipy.stats import multivariate_normal


def sample_direct_MC(dimension):
    # sample in standard normal space, variates are transformed before they are passed to
    # the performance function:
    pi = multivariate_normal(mean=np.zeros(dimension), cov=None)
    x = np.atleast_1d(pi.rvs())
    return {"x": x}



"""
Parameters for a single crystal.
"""
from __future__ import annotations
import copy
from typing_extensions import Self, TypedDict

from hpcflow.sdk.core.parameters import ParameterValue
from hpcflow.sdk.core.utils import get_in_container, set_in_container
from matflow.param_classes.orientations import Orientations


class Perturbation(TypedDict):
    """
    A single perturbation to apply.
    """

    #: Where to apply the perturbation to.
    path: list[str | int]
    #: The multiplicative perturbation to apply.
    multiplicative: float


class SingleCrystalParameters(ParameterValue):
    """
    Parameter relating to the phases in a single crystal.

    Parameters
    ----------
    phases
        The data used to create the bulk phases.
    perturbations
        The perturbations to apply to the phases.
    """

    _typ = "single_crystal_parameters"

    def __init__(
        self,
        phases: dict[str, dict[str, list[float]]],
        perturbations: Perturbation | list[Perturbation] | None = None,
    ):
        self._base = phases
        if perturbations is not None:
            self._perturbations = (
                perturbations if isinstance(perturbations, list) else [perturbations]
            )
        else:
            self._perturbations = []
        # assigned (perturbations applied) on first access
        self._phases: dict[str, dict[str, list[float]]] | None = None

    def __getitem__(self, name: str) -> dict[str, list[float]]:
        """Dict-like retrieval of the parameters for a given phase, with perturbations
        applied."""
        return self.phases[name]

    def to_dict(self):
        out = {k.lstrip("_"): v for k, v in super().to_dict().items() if k != "_phases"}
        out["phases"] = out.pop("base")
        return out

    def as_base(self) -> Self:
        """Return a copy where `base` includes the perturbations."""
        return self.__class__(phases=self.phases)

    @property
    def base(self) -> dict[str, dict[str, list[float]]]:
        """
        The initial data used to create the bulk phases.
        """
        return self._base

    @property
    def phases(self) -> dict[str, dict[str, list[float]]]:
        """
        The data used to create the bulk phases, with perturbations applied.
        """
        if not self._phases:
            phases = copy.deepcopy(self._base)

            perturbations = self._perturbations
            if not isinstance(perturbations, list):
                perturbations = [perturbations]

            for pert_i in perturbations:
                if not pert_i:
                    continue
                scale = pert_i["multiplicative"]
                new_val = get_in_container(phases, pert_i["path"]) * scale
                set_in_container(phases, pert_i["path"], new_val)
            self._phases = phases
        return self._phases

    @property
    def perturbations(self) -> list[Perturbation]:
        """
        The perturbations to apply to the phases.
        """
        return self._perturbations



"""Test `Orientations` and related classes."""

from __future__ import annotations
from pathlib import Path
import sys
import requests
from typing import TYPE_CHECKING
import pytest

from matflow.tests.utils import make_test_data_YAML_workflow

if TYPE_CHECKING:
    from matflow.param_classes.orientations import Orientations


@pytest.mark.xfail(
    condition=sys.platform == "darwin",
    raises=requests.exceptions.HTTPError,
    reason=(
        "GHA MacOS runners use the same IP address, so we get rate limited when "
        "retrieving demo data from GitHub. This happens when testing the "
        "Pyinstaller-built executable, for which we do not include the demo data."
    ),
)
def test_orientations_yaml_init(
    null_config,
    tmp_path: Path,
    orientations_1: Orientations,
    orientations_2: Orientations,
):
    wk = make_test_data_YAML_workflow("define_orientations.yaml", path=tmp_path)
    orientations_t1 = wk.tasks.define_orientations_1.elements[0].inputs.orientations.value
    orientations_t2 = wk.tasks.define_orientations_2.elements[0].inputs.orientations.value
    assert orientations_t1 == orientations_1
    assert orientations_t2 == orientations_2



"""
Miscellaneous utility functions.
"""
from __future__ import annotations
from typing_extensions import TypeIs
import numpy as np
from numpy.typing import NDArray
from pathlib import Path


def _is_list_of_lists(arr: list[float] | list[list[float]]) -> TypeIs[list[list[float]]]:
    return isinstance(arr[0], list)


def masked_array_from_list(
    arr: list[float] | list[list[float]], fill_value: str = "x"
) -> NDArray:
    """Generate a (masked) array from a 1D or 2D list whose elements may contain a fill
    value."""

    is_2D = False
    if _is_list_of_lists(arr):
        is_2D = True
        n_rows = len(arr)
        arr = [item for row in arr for item in row]

    data = np.empty(len(arr))
    mask = np.full(len(arr), False)
    has_mask = False
    for idx, i in enumerate(arr):
        if i == fill_value:
            mask[idx] = True
            has_mask = True
        else:
            data[idx] = i
    if has_mask:
        out = np.ma.masked_array(data, mask=mask)
    else:
        out = data
    if is_2D:
        out = out.reshape(n_rows, -1, order="C")
    return out


def read_numeric_csv_file(
    path: str,
    number: int | None = None,
    start_index: int = 0,
    delimiter: str = " ",
    columns: list[int] | None = None,
    exc_msg: str | None = None,
) -> NDArray:
    """
    Load data from a text file.

    Parameters
    ----------
    path
        Path to the file to load from.
    number
        Number of lines to read from the file.
    start_index
        The line number of the file that the seeds start at.
        Allows skipping headers.
    delimiter
        The delimiter separating values in the file.
        Defaults to space, but commas and tabs are also sensible
        (and correspond to CSV and TSV files respectively).
    columns
        The columns in the file to read from.
        Defaults to reading every column.
    exc_msg:
        ValueError message to return.
    """
    data: list[list[float]] = []
    with Path(path).open("rt") as fh:
        for idx, line in enumerate(fh):
            line = line.strip()
            if not line or idx < start_index:
                continue
            elif len(data) < number if number is not None else True:
                values = line.split(delimiter)
                if idx == start_index:
                    columns = columns or list(range(len(values)))
                data.append([float(values[i]) for i in columns])
    if number is not None and len(data) < number:
        exc_msg = exc_msg or "Not enough lines in the file."
        raise ValueError(exc_msg)

    return np.asarray(data)



from __future__ import annotations
from damask_parse.utils import volume_element_from_2D_microstructure


def generate_volume_element_extrusion(
    microstructure_image: dict,
    depth: int,
    image_axes: list[str],
    homog_label: str,
    phase_label: str,
    phase_label_mapping: dict[str, str],
) -> dict:
    """Extrude a 2D microstructure by a given depth to form a 3D volume element.

    Parameters
    ----------
    microstructure_image : dict
        Dict with the following keys:
            grains : ndarray or nested list of shape (N, M)
                2D map of grain indices.
            orientations : dict
                Dict with the following keys:
                    type: str, "quat"
                    quaternions : ndarray of shape (P, 4) of float
                        Array of P row four-vectors of unit quaternions.
                    unit_cell_alignment : dict
                        Alignment of the unit cell.
            scale : float, optional
            grain_phases: list of int
                Index of phase assigned to each grain
            phase_labels: list of str
                Label of each phase
    depth : int
        By how many voxels the microstructure should be extruded.
    image_axes : list
        Directions along the ndarray axes. Possible values ('x', 'y', 'z')
    homog_label : str
        Homogenization scheme label.
    phase_label : str
        Label of the phase for single phase.
    phase_label_mapping: dict
        Mapping from phase labels in the `microstructure_image` to phase
        labels for created in the VE.

    Returns
    -------
    volume_element : dict
        Dict representation of the volume element.
    """
    volume_element = volume_element_from_2D_microstructure(
        microstructure_image=microstructure_image,
        homog_label=homog_label,
        phase_label=phase_label,
        phase_label_mapping=phase_label_mapping,
        depth=depth,
        image_axes=image_axes,
    )
    volume_element["grid_size"] = tuple(int(i) for i in volume_element["grid_size"])
    volume_element["size"] = tuple(float(i) for i in volume_element["size"])
    return {"volume_element": volume_element}



from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np
from numpy.typing import ArrayLike

try:
    from damask import GeomGrid as grid_cls
except ImportError:
    from damask import Grid as grid_cls
from damask_parse.utils import validate_volume_element, validate_orientations

if TYPE_CHECKING:
    from matflow.param_classes.orientations import Orientations
    from matflow.param_classes.seeds import MicrostructureSeeds


def generate_volume_element_voronoi(
    microstructure_seeds: MicrostructureSeeds,
    VE_grid_size: ArrayLike,
    homog_label: str,
    orientations: Orientations | None,
    scale_morphology: ArrayLike | None,
    scale_update_size: bool,
    periodic: bool,
) -> dict:
    """
    Generate a volume element by Voronoi tessellation from a provided set of seeds.

    Parameters
    ----------
    microstructure_seeds
        Seeds for the crystals in the Voronoi tessellation.
    VE_grid_size
        Volume element grid size
    homog_label
        Homogenization scheme label.
    orientations
        Orientation data for the seeds.
    scale_morphology
        How to scale the morphology.
    scale_update_size
        Whether to update the grid size based on the morphology scaling.
    periodic
        Whether to use periodic boundary conditions on the tessellation.
        Defaults to true.
    """
    grid_obj = grid_cls.from_Voronoi_tessellation(
        cells=np.array(VE_grid_size),
        size=np.array(microstructure_seeds.box_size),
        seeds=np.array(microstructure_seeds.position),
        periodic=periodic,
    )

    if scale_morphology is not None:
        scale_morphology = np.array(scale_morphology)

        original_cells = grid_obj.cells
        new_cells = original_cells * scale_morphology
        grid_scaled = grid_obj.scale(new_cells)

        if scale_update_size:
            original_size = grid_obj.size
            new_size = original_size * scale_morphology
            grid_scaled.size = new_size

        grid_obj = grid_scaled

    # TODO: this needs some more thought.

    if orientations is None:
        orientations = microstructure_seeds.orientations
        assert orientations is not None

    # see `LatticeDirection` enum:
    align_lookup = {
        "A": "a",
        "B": "b",
        "C": "c",
        "A_STAR": "a*",
        "B_STAR": "b*",
        "C_STAR": "c*",
    }
    unit_cell_alignment = {
        "x": align_lookup[orientations.unit_cell_alignment.x.name],
        "y": align_lookup[orientations.unit_cell_alignment.y.name],
        "z": align_lookup[orientations.unit_cell_alignment.z.name],
    }
    type_lookup = {
        "QUATERNION": "quat",
        "EULER": "euler",
    }
    type_ = type_lookup[orientations.representation.type.name]
    oris = {
        "type": type_,
        "unit_cell_alignment": unit_cell_alignment,
    }

    if type_ == "quat":
        quat_order = orientations.representation.quat_order.name.lower().replace("_", "-")
        oris["quaternions"] = np.array(orientations.data)
        oris["quat_component_ordering"] = quat_order
    elif type_ == "euler":
        oris["euler_angles"] = np.array(orientations.data)
        oris["euler_degrees"] = orientations.representation.euler_is_degrees

    oris = validate_orientations(oris)

    num_grains = len(microstructure_seeds.position)
    const_phase_lab = np.array([microstructure_seeds.phase_label])[
        np.zeros(num_grains, dtype=int)
    ]

    ori_idx = np.arange(num_grains)

    volume_element = {
        "size": grid_obj.size.astype(float).tolist(),
        "grid_size": grid_obj.cells.tolist(),
        "orientations": oris,
        "element_material_idx": grid_obj.material,
        "constituent_material_idx": np.arange(num_grains),
        "constituent_material_fraction": np.ones(num_grains),
        "constituent_phase_label": const_phase_lab,
        "constituent_orientation_idx": ori_idx,
        "material_homog": np.full(num_grains, homog_label),
    }
    volume_element = validate_volume_element(volume_element)
    return {"volume_element": volume_element}



from __future__ import annotations
from pathlib import Path

from damask_parse.utils import generate_viz


def viz_damask_HDF5(
    damask_hdf5_file: Path | str,
    damask_viz: dict | list,
    VE_response: dict,
) -> None:
    generate_viz(hdf5_path=damask_hdf5_file, viz_spec=damask_viz, parsed_outs=VE_response)



import pprint
from typing import Optional
import numpy as np
from scipy.stats import norm
from numpy.typing import NDArray


def model(x):
    return np.sum(x)


def get_y_star(p_f, dimension):
    return np.sqrt(dimension) * norm.ppf(1 - p_f)


def system_analysis_toy_model(x: NDArray, dimension: int, target_pf: float):
    """`x` is within the failure domain if the return is greater than zero."""
    x = x[:]  # convert to numpy array
    y_star = get_y_star(target_pf, dimension)
    g_i = model(x) - y_star
    return {"g": g_i}



from PyInstaller.utils.hooks import collect_data_files

hiddenimports = [
    "matflow.data.demo_data_manifest",
    "matflow.data.scripts",
    "matflow.data.template_components",
    "matflow.data.workflows",
    "matflow.param_classes",
    "matflow.tests.data",
]

py_include_kwargs = dict(include_py_files=True, excludes=("**/__pycache__",))
datas = (
    collect_data_files("matflow.data.demo_data_manifest")
    + collect_data_files("matflow.data.scripts", **py_include_kwargs)
    + collect_data_files("matflow.data.template_components")
    + collect_data_files("matflow.data.workflows")
    + collect_data_files("matflow.tests", **py_include_kwargs)
    + collect_data_files("matflow.tests.data")
)



"""Utility functions to assist testing."""

from functools import partial

from hpcflow.sdk.core.test_utils import (
    make_test_data_YAML_workflow,
    make_test_data_YAML_workflow_template,
)

import matflow as mf


make_test_data_YAML_workflow = partial(
    make_test_data_YAML_workflow, app=mf, pkg="matflow.tests.data"
)

make_test_data_YAML_workflow_template = partial(
    make_test_data_YAML_workflow_template, app=mf, pkg="matflow.tests.data"
)
